import asyncio
import enum
import functools
import json
import sys
import textwrap
import traceback
from typing import Sequence

from click import echo, style
from tqdm import tqdm

from ..exceptions import BackendAPIError

__all__ = (
    "PrintStatus",
    "print_pretty",
    "print_info",
    "print_wait",
    "print_done",
    "print_warn",
    "print_fail",
    "print_error",
    "show_warning",
)


class PrintStatus(enum.Enum):
    NONE = 0
    WAITING = 1
    DONE = 2
    FAILED = 3
    WARNING = 4


def bold(text: str) -> str:
    """
    Wraps the given text with bold enable/disable ANSI sequences.
    """
    return style(text, bold=True, reset=False) + style("", bold=False, reset=False)


def underline(text: str) -> str:
    return style(text, underline=True, reset=False) + style("", underline=False, reset=False)


def inverse(text: str) -> str:
    return style(text, reverse=True, reset=False) + style("", reverse=False, reset=False)


def italic(text: str) -> str:
    return "\x1b[3m" + text + "\x1b[23m"


def format_pretty(msg, status=PrintStatus.NONE, colored=True):
    if status == PrintStatus.NONE:
        indicator = style("\u2219", fg="bright_cyan", reset=False)
    elif status == PrintStatus.WAITING:
        indicator = style("\u22ef", fg="bright_yellow", reset=False)
    elif status == PrintStatus.DONE:
        indicator = style("\u2714", fg="bright_green", reset=False)
    elif status == PrintStatus.FAILED:
        indicator = style("\u2718", fg="bright_red", reset=False)
    elif status == PrintStatus.WARNING:
        indicator = style("\u2219", fg="yellow", reset=False)
    else:
        raise ValueError
    return style(indicator + textwrap.indent(msg, "  ")[1:], reset=True)


format_info = functools.partial(format_pretty, status=PrintStatus.NONE)
format_wait = functools.partial(format_pretty, status=PrintStatus.WAITING)
format_done = functools.partial(format_pretty, status=PrintStatus.DONE)
format_fail = functools.partial(format_pretty, status=PrintStatus.FAILED)
format_warn = functools.partial(format_pretty, status=PrintStatus.WARNING)


def print_pretty(msg, *, status=PrintStatus.NONE, file=None):
    if file is None:
        file = sys.stderr
    if status == PrintStatus.NONE:
        indicator = style("\u2219", fg="bright_cyan", reset=False)
    elif status == PrintStatus.WAITING:
        assert "\n" not in msg, "Waiting message must be a single line."
        indicator = style("\u22ef", fg="bright_yellow", reset=False)
    elif status == PrintStatus.DONE:
        indicator = style("\u2713", fg="bright_green", reset=False)
    elif status == PrintStatus.FAILED:
        indicator = style("\u2717", fg="bright_red", reset=False)
    elif status == PrintStatus.WARNING:
        indicator = style("\u2219", fg="yellow", reset=False)
    else:
        raise ValueError
    echo("\x1b[2K", nl=False, file=file)
    text = textwrap.indent(msg, "  ")
    text = style(indicator + text[1:], reset=True)
    echo("{0}\r".format(text), nl=False, file=file)
    file.flush()
    if status != PrintStatus.WAITING:
        echo("", file=file)


print_info = functools.partial(print_pretty, status=PrintStatus.NONE)
print_wait = functools.partial(print_pretty, status=PrintStatus.WAITING)
print_done = functools.partial(print_pretty, status=PrintStatus.DONE)
print_fail = functools.partial(print_pretty, status=PrintStatus.FAILED)
print_warn = functools.partial(print_pretty, status=PrintStatus.WARNING)


def _format_gql_path(items: Sequence[str | int]) -> str:
    pieces = []
    for item in items:
        match item:
            case int():
                pieces.append(f"[{item}]")
            case _:
                pieces.append(f".{str(item)}")
    return "".join(pieces)[1:]  # strip first dot


def format_error(exc: Exception):
    if isinstance(exc, BackendAPIError):
        yield "{0}: {1} {2}\n".format(exc.__class__.__name__, exc.status, exc.reason)
        yield "{0[title]}".format(exc.data)
        if exc.data["type"].endswith("/too-many-sessions-matched"):
            matches = exc.data["data"].get("matches", [])
            if matches:
                yield "\nCandidates (up to 10 recent entries):\n"
            for item in matches:
                yield f"- {item['id']} ({item['name']}, {item['status']})\n"
        elif exc.data["type"].endswith("/session-already-exists"):
            existing_session_id = exc.data["data"].get("existingSessionId", None)
            if existing_session_id is not None:
                yield f"\n- Existing session ID: {existing_session_id}"
        elif exc.data["type"].endswith("/invalid-api-params"):
            general_error_msg = exc.data.get("msg", None)
            if general_error_msg is not None:
                yield f"\n- {general_error_msg}"
            per_field_errors = exc.data.get("data", None)
            if per_field_errors:
                yield "\n"
                yield json.dumps(per_field_errors, indent=2)
        else:
            if exc.data["type"].endswith("/graphql-error"):
                yield "\n\u279c Message:\n"
                for err_item in exc.data.get("data", []):
                    yield f"{err_item['message']}"
                    if err_path := err_item.get("path"):
                        yield f" (path: {_format_gql_path(err_path)})"
                    yield "\n"
            else:
                other_details = exc.data.get("msg", None)
                if other_details:
                    yield "\n\u279c Message: "
                    yield str(other_details)
                other_data = exc.data.get("data", None)
                if other_data:
                    yield "\n\u279c Data: "
                    yield repr(other_data)
        agent_details = exc.data.get("agent-details", None)
        if agent_details is not None:
            yield "\n\u279c This is an agent-side error. "
            yield "Check the agent status or ask the administrator for help."
            agent_exc = agent_details.get("exception", None)
            if agent_exc is not None:
                yield "\n\u279c " + str(agent_exc)
            desc = agent_details.get("title", None)
            if desc is not None:
                yield "\n\u279c " + str(desc)
        content = exc.data.get("content", None)
        if content:
            yield "\n" + content
    else:
        args = exc.args if exc.args else [""]
        yield f"{exc.__class__.__name__}: {args[0]}\n"
        yield "\n".join(map(str, args[1:]))
        yield ("*** Traceback ***\n" + "".join(traceback.format_tb(exc.__traceback__)).strip())


def print_error(exc: Exception, *, file=None):
    if file is None:
        file = sys.stderr
    indicator = style("\u2718", fg="bright_red", reset=False)
    if file.isatty():
        echo("\x1b[2K", nl=False, file=file)
    text = "".join(format_error(exc))
    text = textwrap.indent(text, "  ")
    text = style(indicator + text[1:], reset=True)
    echo("{0}\r".format(text), nl=False, file=file)
    echo("", file=file)
    file.flush()


def show_warning(message, category, filename, lineno, file=None, line=None):
    echo(
        "{0}: {1}".format(
            style(str(category.__name__), fg="yellow", bold=True),
            style(str(message), fg="yellow"),
        ),
        file=file,
    )


class Spinner:
    def __init__(self, msg: str, delay: float = 0.3):
        self.msg = msg
        self.task = None
        self.delay = delay

    async def __aenter__(self):
        self.run()
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        await self.stop()

    def run(self):
        self.task = asyncio.create_task(self.spin())

    async def spin(self):
        try:
            while True:
                for char in "|/-\\":
                    print_wait("{} {}".format(self.msg, char))
                    await asyncio.sleep(self.delay)
        except asyncio.CancelledError:
            pass

    async def stop(self):
        self.task.cancel()
        await self.task


class ProgressViewer:
    """

    A context manager that displays a spinner and a tqdm progress bar.

    It shows the spinner until it is switched explicitly to the tqdm progress bar.

    Usage:

    ```
    async with ProgressViewer("Waiting...") as viewer:
        for i in range(10):
            await asyncio.sleep(0.2)
        tqdm = await viewer.to_tqdm()
        tqdm.total = 10
        for i in range(10):
            await asyncio.sleep(0.2)
            tqdm.update(1)
    ```
    """

    def __init__(self, spinner_msg: str = "", delay: float = 0.3) -> None:
        self.spinner = Spinner(spinner_msg, delay)
        self.tqdm = None

    async def __aenter__(self):
        self.spinner.run()
        return self

    async def __aexit__(self, exc_type, exc_value, traceback) -> None:
        await self.spinner.stop()
        if self.tqdm:
            self.tqdm.close()

    async def to_tqdm(self, unit: str = "it") -> tqdm:
        await self.spinner.stop()
        self.tqdm = tqdm(total=0, unit=unit)
        return self.tqdm



import sys
from pathlib import Path

import click
import humanize

from ai.backend.cli.main import main
from ai.backend.cli.params import (
    ByteSizeParamCheckType,
    ByteSizeParamType,
    CommaSeparatedKVListParamType,
)
from ai.backend.cli.types import ExitCode
from ai.backend.client.config import DEFAULT_CHUNK_SIZE, APIConfig
from ai.backend.client.output.fields import vfolder_fields
from ai.backend.client.session import Session

from ..exceptions import BackendAPIError
from ..output.types import FieldSpec
from .extensions import pass_ctx_obj
from .pretty import print_done
from .types import CLIContext


@main.group()
def model():
    """Set of model operations"""


@model.command()
@pass_ctx_obj
@click.option("--filter", "filter_", default=None, help="Set the query filter expression.")
@click.option("--order", default=None, help="Set the query ordering expression.")
@click.option("--offset", default=0, help="The index of the current page start for pagination.")
@click.option("--limit", type=int, default=None, help="The page size for pagination.")
def list(ctx: CLIContext, filter_, order, offset, limit):
    """
    List the models.
    """

    with Session() as session:
        try:
            fetch_func = lambda pg_offset, pg_size: session.Model.paginated_list(
                page_offset=pg_offset,
                page_size=pg_size,
                filter=filter_,
                order=order,
            )
            ctx.output.print_paginated_list(
                fetch_func,
                initial_page_offset=offset,
                page_size=limit,
            )
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(ExitCode.FAILURE)


@model.command()
@pass_ctx_obj
@click.argument("model_name", metavar="MODEL", type=str)
def info(ctx: CLIContext, model_name):
    """
    Display the detail of a model with its backing storage vfolder.

    \b
    MODEL: The model name
    """

    with Session() as session:
        try:
            result = session.Model(model_name).info()
            ctx.output.print_item(
                result,
                [
                    vfolder_fields["id"],
                    vfolder_fields["name"],
                    FieldSpec("versions"),
                ],
            )
        except BackendAPIError as e:
            ctx.output.print_fail(
                "Not a valid model storage. "
                "There is no directory named `versions` under the model storage "
                "or model storage not found."
            )
            ctx.output.print_error(e)
            sys.exit(ExitCode.FAILURE)
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(ExitCode.FAILURE)


@model.command()
@pass_ctx_obj
@click.argument("name", type=str)
@click.argument("host", type=str, default=None)
@click.option(
    "-g",
    "--group",
    metavar="GROUP",
    type=str,
    default=None,
    help="Group ID or NAME.",
)
@click.option(
    "--unmanaged",
    "host_path",
    is_flag=True,
    help=(
        "Treats HOST as a mount point of unmanaged model. "
        "This option can only be used by Admin or Superadmin."
    ),
)
@click.option(
    "-p",
    "--permission",
    metavar="PERMISSION",
    type=str,
    default="rw",
    help=(
        "Folder's innate permission. "
        'Group folders can be shared as read-only by setting this option to "ro". '
        "Invited folders override this setting by its own invitation permission."
    ),
)
@click.option(
    "-q",
    "--quota",
    metavar="QUOTA",
    type=ByteSizeParamCheckType(),
    default="0",
    help=(
        "Quota of the virtual folder. "
        "(Use 'm' for megabytes, 'g' for gigabytes, and etc.) "
        "Default is maximum amount possible."
    ),
)
@click.option(
    "--cloneable",
    "--allow-clone",
    is_flag=True,
    help="Allows the virtual folder to be cloned by users.",
)
def create(ctx: CLIContext, name, host, group, host_path, permission, quota, cloneable):
    """
    Create a new model with the given configuration.

    \b
    NAME: Name of a model.
    HOST: Name of a virtual folder host in which the model will be created.
    """
    with Session() as session:
        try:
            if host_path:
                result = session.Model.create(
                    name=name,
                    host=host,
                    unmanaged_path=host_path,
                    group=group,
                    permission=permission,
                    quota=quota,
                    cloneable=cloneable,
                )
            else:
                result = session.Model.create(
                    name=name,
                    host=host,
                    group=group,
                    permission=permission,
                    quota=quota,
                    cloneable=cloneable,
                )
            print_done("Created the model.")
            ctx.output.print_item(result, [FieldSpec("id"), FieldSpec("name")])
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(ExitCode.FAILURE)


@model.command()
@pass_ctx_obj
@click.argument("model_name", metavar="MODEL", type=str)
def rm(ctx: CLIContext, model_name):
    """
    Remove the given model.

    \b
    MODEL: The model ID.
    """

    with Session() as session:
        try:
            serving = session.Model(model_name)
            serving.delete()
            print_done("Model deleted.")
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(ExitCode.FAILURE)


@model.command()
@pass_ctx_obj
@click.argument("model_name", metavar="MODEL", type=str)
@click.argument("filenames", type=Path, nargs=-1)
@click.argument("model_version", metavar="MODEL_VER", type=str)
@click.option(
    "-b",
    "--base-dir",
    type=Path,
    default=None,
    help=(
        "The local parent directory which contains the file to be uploaded.  "
        "[default: current working directry]"
    ),
)
@click.option(
    "--chunk-size",
    type=ByteSizeParamType(),
    default=humanize.naturalsize(DEFAULT_CHUNK_SIZE, binary=True, gnu=True),
    help=(
        'Transfer the file with the given chunk size with binary suffixes (e.g., "16m"). '
        "Set this between 8 to 64 megabytes for high-speed disks (e.g., SSD RAID) "
        "and networks (e.g., 40 GbE) for the maximum throughput."
    ),
)
@click.option(
    "--override-storage-proxy",
    type=CommaSeparatedKVListParamType(),
    default=None,
    help=(
        "Overrides storage proxy address. "
        'The value must shape like "X1=Y1,X2=Y2...". '
        "Each Yn address must at least include the IP address "
        "or the hostname and may include the protocol part and the port number to replace."
    ),
)
def upload(
    ctx: CLIContext,
    model_name,
    filenames,
    model_version,
    base_dir,
    chunk_size,
    override_storage_proxy,
):
    """
    Upload a file to the model as the given version.
    The files with the same names will be overwirtten.

    \b
    MODEL: The model ID
    FILENAMES: The uploaded files paths relative to the current working directory
    """
    with Session() as session:
        try:
            session.VFolder(model_name).upload(
                filenames,
                dst_dir=Path("versions", model_version),
                basedir=base_dir,
                chunk_size=chunk_size,
                show_progress=True,
                address_map=(
                    override_storage_proxy or APIConfig.DEFAULTS["storage_proxy_address_map"]
                ),
            )
            print_done("Done.")
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(ExitCode.FAILURE)


@model.command()
@pass_ctx_obj
@click.argument("model_name", type=str)
@click.argument("filenames", type=Path, nargs=-1)
@click.argument("model_version", metavar="MODEL_VER", type=str)
@click.option(
    "-b",
    "--base-dir",
    type=Path,
    default=None,
    help=(
        "The local parent directory which will contain the downloaded file.  "
        "[default: current working directry]"
    ),
)
@click.option(
    "--chunk-size",
    type=ByteSizeParamType(),
    default=humanize.naturalsize(DEFAULT_CHUNK_SIZE, binary=True, gnu=True),
    help=(
        'Transfer the file with the given chunk size with binary suffixes (e.g., "16m"). '
        "Set this between 8 to 64 megabytes for high-speed disks (e.g., SSD RAID) "
        "and networks (e.g., 40 GbE) for the maximum throughput."
    ),
)
@click.option(
    "--override-storage-proxy",
    type=CommaSeparatedKVListParamType(),
    default=None,
    help=(
        "Overrides storage proxy address. "
        'The value must shape like "X1=Y1,X2=Y2...". '
        "Each Yn address must at least include the IP address "
        "or the hostname and may include the protocol part and the port number to replace."
    ),
)
@click.option(
    "--max-retries",
    type=int,
    default=20,
    help="Maximum retry attempt when any failure occurs.",
)
def download(
    ctx: CLIContext,
    model_name,
    filenames,
    model_version,
    base_dir,
    chunk_size,
    override_storage_proxy,
    max_retries,
):
    """
    Download a file from the model storage.
    The files with the same names will be overwirtten.

    \b
    MODEL: The model ID
    FILENAMES: The file paths in the model storage vfolder to download to the current working
               directory.
    """
    with Session() as session:
        try:
            session.VFolder(model_name).download(
                filenames,
                dst_dir=Path("versions", model_version),
                basedir=base_dir,
                chunk_size=chunk_size,
                show_progress=True,
                address_map=(
                    override_storage_proxy or APIConfig.DEFAULTS["storage_proxy_address_map"]
                ),
                max_retries=max_retries,
            )
            print_done("Done.")
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(ExitCode.FAILURE)



from __future__ import annotations

import sys
import uuid
from collections.abc import Iterable, Sequence

import click

from ai.backend.cli.interaction import ask_yn
from ai.backend.cli.params import BoolExprType, CommaSeparatedListType, OptionalType
from ai.backend.cli.types import ExitCode, Undefined, undefined
from ai.backend.client.exceptions import BackendAPIError
from ai.backend.client.output.fields import user_fields
from ai.backend.client.session import Session

from ..extensions import pass_ctx_obj
from ..pretty import print_fail, print_info
from ..types import CLIContext
from . import admin


@admin.group()
def user() -> None:
    """
    User administration commands.
    """


@user.command()
@pass_ctx_obj
@click.option("-e", "--email", type=str, default=None, help="Email of a user to display.")
def info(ctx: CLIContext, email: str) -> None:
    """
    Show the information about the given user by email. If email is not given,
    requester's information will be displayed.
    """
    fields = [
        user_fields["uuid"],
        user_fields["username"],
        user_fields["role"],
        user_fields["email"],
        user_fields["full_name"],
        user_fields["description"],
        user_fields["need_password_change"],
        user_fields["status"],
        user_fields["status_info"],
        user_fields["created_at"],
        user_fields["domain_name"],
        user_fields["groups"],
        user_fields["allowed_client_ip"],
        user_fields["sudo_session_enabled"],
        user_fields["main_access_key"],
    ]
    with Session() as session:
        try:
            item = session.User.detail(email=email, fields=fields)
            ctx.output.print_item(item, fields=fields)
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(ExitCode.FAILURE)


@user.command()
@pass_ctx_obj
@click.option(
    "-s",
    "--status",
    type=str,
    default=None,
    help="Filter users in a specific state (active, inactive, deleted, before-verification).",
)
@click.option(
    "-g",
    "--group",
    type=str,
    default=None,
    help="""
    Filter by group ID.

    \b
    EXAMPLE
        --group "$(backend.ai admin group list | grep 'example-group-name' | awk '{print $1}')"
    """,
)
@click.option(
    "--filter",
    "filter_",
    default=None,
    help="""
    Set the query filter expression.

    \b
    COLUMNS
        uuid, username, role, email, full_name, need_password_change,
        status, status_info, created_at, modified_at, domain_name, allowed_client_ip

    \b
    OPERATORS
        Binary Operators: ==, !=, <, <=, >, >=, is, isnot, like, ilike(case-insensitive), in, contains
        Condition Operators: &, |
        Special Symbol: % (wildcard for like and ilike operators)

    \b
    EXAMPLE QUERIES
        --filter 'status == "ACTIVE" & role in ["ADMIN", "SUPERADMIN"]'
        --filter 'created_at >= "2021-01-01" & created_at < "2023-01-01"'
        --filter 'email ilike "%@example.com"'
    """,
)
@click.option(
    "--order",
    default=None,
    help="""
    Set the query ordering expression.

    \b
    COLUMNS
        uuid, username, role, email, full_name, need_password_change,
        status, status_info, created_at, modified_at, domain_name

    \b
    OPTIONS
        ascending order (default): (+)column_name
        descending order: -column_name

    \b
    EXAMPLE
        --order 'uuid'
        --order '+uuid'
        --order '-created_at'
    """,
)
@click.option("--offset", default=0, help="The index of the current page start for pagination.")
@click.option("--limit", type=int, default=None, help="The page size for pagination.")
def list(ctx: CLIContext, status, group, filter_, order, offset, limit) -> None:
    """
    List users.
    (admin privilege required)
    """
    fields = [
        user_fields["uuid"],
        user_fields["username"],
        user_fields["role"],
        user_fields["email"],
        user_fields["full_name"],
        user_fields["need_password_change"],
        user_fields["status"],
        user_fields["status_info"],
        user_fields["created_at"],
        user_fields["domain_name"],
        user_fields["groups"],
        user_fields["allowed_client_ip"],
        user_fields["sudo_session_enabled"],
        user_fields["main_access_key"],
    ]
    try:
        with Session() as session:
            fetch_func = lambda pg_offset, pg_size: session.User.paginated_list(
                status,
                group,
                fields=fields,
                page_offset=pg_offset,
                page_size=pg_size,
                filter=filter_,
                order=order,
            )
            ctx.output.print_paginated_list(
                fetch_func,
                initial_page_offset=offset,
                page_size=limit,
            )
    except Exception as e:
        ctx.output.print_error(e)
        sys.exit(ExitCode.FAILURE)


@user.command(aliases=["create"])
@pass_ctx_obj
@click.argument("domain_name", type=str, metavar="DOMAIN_NAME")
@click.argument("email", type=str, metavar="EMAIL")
@click.argument("password", type=str, metavar="PASSWORD")
@click.option("-u", "--username", type=str, default="", help="Username.")
@click.option("-n", "--full-name", type=str, default="", help="Full name.")
@click.option(
    "-r",
    "--role",
    type=str,
    default="user",
    help="Role of the user. One of (admin, user, monitor).",
)
@click.option(
    "-s",
    "--status",
    type=str,
    default="active",
    help="Account status. One of (active, inactive, deleted, before-verification).",
)
@click.option(
    "--need-password-change",
    is_flag=True,
    help=(
        "Flag indicate that user needs to change password. "
        "Useful when admin manually create password."
    ),
)
@click.option(
    "--allowed-ip",
    type=CommaSeparatedListType(),
    default=None,
    help=(
        "Allowed client IP. IPv4 and IPv6 are allowed. CIDR type is recommended. "
        '(e.g., --allowed-ip "127.0.0.1","127.0.0.2",...)'
    ),
)
@click.option(
    "--desc",
    "--description",
    "description",
    type=str,
    default="",
    help="Description of the user.",
)
@click.option(
    "--sudo-session-enabled",
    is_flag=True,
    default=False,
    help=(
        "Enable passwordless sudo for a user inside a compute session. "
        "Note that this feature does not automatically install sudo for the session."
    ),
)
@click.option(
    "--cuid",
    "--container-uid",
    "container_uid",
    type=int,
    default=None,
    help="The user ID (UID) that will be assigned to all processes running inside containers created by this user.",
)
@click.option(
    "--cgid",
    "--container-main-gid",
    "container_main_gid",
    type=int,
    default=None,
    help="The primary group ID (GID) that will be assigned to all processes running inside containers created by this user.",
)
@click.option(
    "--cgids",
    "--container-gids",
    "container_gids",
    type=CommaSeparatedListType(int),
    default=None,
    help=(
        "Supplementary group IDs that will be assigned to all processes running inside containers created by this user. "
        "(e.g., --cgids 1001,1002,1003)"
    ),
)
@click.option(
    "-g",
    "--group",
    "--groups",
    "groups",
    type=CommaSeparatedListType(),
    default="default",
    help='Add to the groups. If unspecified, it will use the "default" group.',
)
def add(
    ctx: CLIContext,
    domain_name: str,
    email: str,
    password: str,
    username: str,
    full_name: str,
    role: str,
    status: str,
    need_password_change: bool,
    allowed_ip: str | None,
    description: str,
    sudo_session_enabled: bool,
    container_uid: int | None,
    container_main_gid: int | None,
    container_gids: Iterable[int] | None,
    groups: Iterable[str],
):
    """
    Create a new user. As the user must belong to a domain,
    you should provide DOMAIN_NAME explicitly.

    \b
    DOMAIN_NAME: Name of the domain where new user belongs to.
    EMAIL: Email of new user.
    PASSWORD: Password of new user.
    """
    with Session() as session:
        try:
            group_ids = []
            # Resolve the group names to IDs if necessary.
            for group_ref in groups:
                try:
                    uuid.UUID(group_ref)
                    group_ids.append(group_ref)
                except ValueError:
                    data = session.Group.from_name(
                        group_ref,
                        domain_name=domain_name,
                    )
                    if not data:
                        # Either domain_name or group_ref may be invalid.
                        raise ValueError(
                            f"Cannot find the group {group_ref!r} in the domain {domain_name!r}"
                        )
                    group_ids.append(data[0]["id"])
            data = session.User.create(
                domain_name,
                email,
                password,
                username=username,
                full_name=full_name,
                role=role,
                status=status,
                need_password_change=need_password_change,
                allowed_client_ip=allowed_ip,
                group_ids=group_ids,
                description=description,
                sudo_session_enabled=sudo_session_enabled,
                container_uid=container_uid,
                container_main_gid=container_main_gid,
                container_gids=container_gids,
                fields=(
                    user_fields["domain_name"],
                    user_fields["email"],
                    user_fields["username"],
                    user_fields["uuid"],
                    user_fields["groups"],
                ),
            )
        except BackendAPIError as e:
            ctx.output.print_mutation_error(
                e,
                item_name="user",
                action_name="add",
            )
            sys.exit(ExitCode.FAILURE)
        except ValueError as e:
            ctx.output.print_fail(str(e))
            sys.exit(ExitCode.FAILURE)
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(ExitCode.FAILURE)

        if not data["ok"]:
            ctx.output.print_mutation_error(
                msg=data["msg"],
                item_name="user",
                action_name="add",
            )
            sys.exit(ExitCode.FAILURE)

        # post-process group names
        data["user"]["groups"] = ", ".join(item["name"] for item in data["user"]["groups"])
        ctx.output.print_mutation_result(
            data,
            item_name="user",
        )


@user.command()
@pass_ctx_obj
@click.argument("email", type=str, metavar="EMAIL")
@click.option(
    "-p",
    "--password",
    type=OptionalType(str),
    default=undefined,
    help="Password.",
)
@click.option(
    "-u",
    "--username",
    type=OptionalType(str),
    default=undefined,
    help="Username.",
)
@click.option(
    "-n",
    "--full-name",
    type=OptionalType(str),
    default=undefined,
    help="Full name.",
)
@click.option(
    "-d",
    "--domain-name",
    type=OptionalType(str),
    default=undefined,
    help="Domain name.",
)
@click.option(
    "-r",
    "--role",
    type=OptionalType(str),
    default=undefined,
    help="Role of the user. One of (admin, user, monitor).",
)
@click.option(
    "-s",
    "--status",
    type=OptionalType(str),
    default=undefined,
    help="Account status. One of (active, inactive, deleted, before-verification).",
)
@click.option(
    "--need-password-change",
    type=OptionalType(BoolExprType),
    default=undefined,
    help=(
        "Flag indicate that user needs to change password. "
        "Useful when admin manually create password."
    ),
)
@click.option(
    "--allowed-ip",
    type=OptionalType(CommaSeparatedListType),
    default=undefined,
    help=(
        "Allowed client IP. IPv4 and IPv6 are allowed. CIDR type is recommended. "
        '(e.g., --allowed-ip "127.0.0.1","127.0.0.2",...)'
    ),
)
@click.option(
    "--description", type=OptionalType(str), default=undefined, help="Description of the user."
)
@click.option(
    "--sudo-session-enabled",
    type=OptionalType(BoolExprType),
    default=undefined,
    help=(
        "Enable passwordless sudo for a user inside a compute session. "
        "Note that this feature does not automatically install sudo for the session."
    ),
)
@click.option(
    "--main-access-key",
    type=OptionalType(str),
    default=undefined,
    help="Set main access key which works as default.",
)
@click.option(
    "--cuid",
    "--container-uid",
    "container_uid",
    type=OptionalType(int),
    default=undefined,
    help="The user ID (UID) that will be assigned to all processes running inside containers created by this user.",
)
@click.option(
    "--unset-cuid",
    "--unset-container-uid",
    "unset_container_uid",
    is_flag=True,
    default=False,
    help="Unset the user's container UID.",
)
@click.option(
    "--cgid",
    "--container-main-gid",
    "container_main_gid",
    type=OptionalType(int),
    default=undefined,
    help="The primary group ID (GID) that will be assigned to all processes running inside containers created by this user.",
)
@click.option(
    "--unset-cgid",
    "--unset-container-main-gid",
    "unset_container_main_gid",
    is_flag=True,
    default=False,
    help="Unset the user's container primary GID.",
)
@click.option(
    "--cgids",
    "--container-gids",
    "container_gids",
    type=OptionalType(CommaSeparatedListType(int)),
    default=undefined,
    help=(
        "Supplementary group IDs that will be assigned to all processes running inside containers created by this user. "
        "(e.g., --cgids 1001,1002,1003)"
    ),
)
@click.option(
    "--unset-cgids",
    "--unset-container-gids",
    "unset_container_gids",
    is_flag=True,
    default=False,
    help="Unset the user's container supplementary group IDs.",
)
def update(
    ctx: CLIContext,
    email: str,
    password: str | Undefined,
    username: str | Undefined,
    full_name: str | Undefined,
    domain_name: str | Undefined,
    role: str | Undefined,
    status: str | Undefined,
    need_password_change: bool | Undefined,
    allowed_ip: Sequence[str] | Undefined,
    description: str | Undefined,
    sudo_session_enabled: bool | Undefined,
    main_access_key: str | Undefined,
    container_uid: int | Undefined,
    unset_container_uid: bool,
    container_main_gid: int | Undefined,
    unset_container_main_gid: bool,
    container_gids: Iterable[int] | Undefined,
    unset_container_gids: bool,
):
    """
    Update an existing user.

    \b
    EMAIL: Email of user to update.
    """

    def validate_input[T_Input](value: T_Input, null_flag: bool, field_name: str) -> T_Input | None:
        _val: T_Input | None = value
        if null_flag:
            if value is not undefined:
                print_fail(
                    f"Cannot set both --{field_name} and --unset-{field_name}. "
                    f"Use --{field_name} to set a value or --unset-{field_name} to set null, but not both."
                )
                sys.exit(ExitCode.FAILURE)
            _val = None
        return _val

    cuid = validate_input(container_uid, unset_container_uid, "cuid")
    cgid = validate_input(container_main_gid, unset_container_main_gid, "cgid")
    cgids = validate_input(container_gids, unset_container_gids, "cgids")

    with Session() as session:
        try:
            data = session.User.update(
                email,
                password=password,
                username=username,
                full_name=full_name,
                domain_name=domain_name,
                role=role,
                status=status,
                need_password_change=need_password_change,
                allowed_client_ip=allowed_ip,
                description=description,
                sudo_session_enabled=sudo_session_enabled,
                main_access_key=main_access_key,
                container_uid=cuid,
                container_main_gid=cgid,
                container_gids=cgids,
            )
        except Exception as e:
            ctx.output.print_mutation_error(
                e,
                item_name="user",
                action_name="update",
            )
            sys.exit(ExitCode.FAILURE)
        if not data["ok"]:
            ctx.output.print_mutation_error(
                msg=data["msg"],
                item_name="user",
                action_name="update",
            )
            sys.exit(ExitCode.FAILURE)
        ctx.output.print_mutation_result(
            data,
            extra_info={
                "email": email,
            },
        )


@user.command()
@pass_ctx_obj
@click.argument("email", type=str, metavar="EMAIL")
def delete(ctx: CLIContext, email):
    """
    Inactivate an existing user.

    \b
    EMAIL: Email of user to inactivate.
    """
    with Session() as session:
        try:
            data = session.User.delete(email)
        except Exception as e:
            ctx.output.print_mutation_error(
                e,
                item_name="user",
                action_name="deletion",
            )
            sys.exit(ExitCode.FAILURE)
        if not data["ok"]:
            ctx.output.print_mutation_error(
                msg=data["msg"],
                item_name="user",
                action_name="deletion",
            )
            sys.exit(ExitCode.FAILURE)
        ctx.output.print_mutation_result(
            data,
            extra_info={
                "email": email,
            },
        )


@user.command()
@pass_ctx_obj
@click.argument("email", type=str, metavar="EMAIL")
@click.option(
    "--purge-shared-vfolders",
    is_flag=True,
    default=False,
    help=(
        "Delete user's all virtual folders. "
        "If False, shared folders will not be deleted "
        "and migrated the ownership to the requested admin."
    ),
)
@click.option(
    "--delegate-endpoint-ownership",
    is_flag=True,
    default=False,
    help=(
        "Delegate user's all alive endpoints. "
        "If True, user's endpoints will not be deleted "
        "and delegate the ownership to the requested admin."
    ),
)
def purge(ctx: CLIContext, email, purge_shared_vfolders, delegate_endpoint_ownership):
    """
    Delete an existing user. This action cannot be undone.

    \b
    NAME: Name of a domain to delete.
    """
    with Session() as session:
        try:
            if not ask_yn():
                print_info("Cancelled")
                sys.exit(ExitCode.FAILURE)
            data = session.User.purge(email, purge_shared_vfolders, delegate_endpoint_ownership)
        except Exception as e:
            ctx.output.print_mutation_error(
                e,
                item_name="user",
                action_name="purge",
            )
            sys.exit(ExitCode.FAILURE)
        if not data["ok"]:
            ctx.output.print_mutation_error(
                msg=data["msg"],
                item_name="user",
                action_name="purge",
            )
            sys.exit(ExitCode.FAILURE)
        ctx.output.print_mutation_result(
            data,
            extra_info={
                "email": email,
            },
        )
